<?php
/**
 * ownCloud
 *
 * @author Juan Pablo Villafañez Ramos <jvillafanez@owncloud.com>
 * @author Jesus Macias Portela <jesus@owncloud.com>
 * @author Jörn Friedrich Dreyer <jfd@owncloud.com>
 * @copyright (C) 2016 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

/**
 * REQUIREMENTS:
 * - libsmbclient and libsmbclient-dev packages installed
 * - libsmbclient-php bindings installed and libsmbclient.so active for php
 * - HOME variable accessible for php (due to a bug in libsmbclient-php)
 */

namespace OCA\windows_network_drive\lib;

use Icewind\SMB\Exception\AlreadyExistsException;
use Icewind\SMB\Exception\Exception;
use Icewind\SMB\Exception\ForbiddenException;
use Icewind\SMB\Exception\NotFoundException;
use Icewind\SMB\NativeFileInfo;
use Icewind\SMB\NativeServer;
use Icewind\SMB\NativeShare;
use Icewind\Streams\CallbackWrapper;
use OC\Cache\CappedMemoryCache;
use OC\Files\Cache\Watcher;
use OC\Files\Filesystem;
use OC\Files\Storage\Common;
use OC\Files\Stream\Dir;
use OCP\Files\StorageNotAvailableException;
use OCP\Util;

/**
 * FS class to connect to windows network drives
 */
class WND extends Common {

    /** @var string Password for the connection */
    private $password;

    /** @var string User to connect */
    private $user;

    /** @var string Host where the connection should be made */
    private $host;

    /** @var string The folder that will act as root */
    private $root;

    /** @var string The share name that we'll use */
    private $shareName;

    /** @var bool is sharing active for this FS? */
    private $isSharingActive = false;

    /** @var CappedMemoryCache a cache for the dir call */
    private $dirCache;

    /** @var CappedMemoryCache a cache for the stat calls */
    private $statCache;

    /** @var CappedMemoryCache a cache for attributes */
    private $xattrCache;

    /** @var NativeServer Reference to a SMB server (check 3rdparty native server) */
    private $server;

    /** @var NativeShare Reference to a SMB share (check 3rdparty native share) */
    private $share;

    /** @var $int */
    private $credentialsType;

    /** @var $bool */
    private $isInitialized = false;

    /** @var $bool */
    private $connectionTested = false;

    /** @var WNDNotifier  */
    private $wndNotifier;

    /**
     * Construct the object. The required params are: "host", "user", "password" and "share".
     * "root" is optional, defaulting to '/'
     *
     * @param array $params an array with the required params
     * @throws \Exception on invalid parameters
     */
    public function __construct($params) {
        // if user is null we convert username and password to '' to try anonymous login
        $params = $this->fixConstructParams($params);
        $this->isSharingActive = $params['oc_sharing'];
        if (!isset($params['host']) || !isset($params['share']) || !isset($params['credentialsType'])) {
            throw new \Exception('invalid parameters in the wnd constructor');
        }

        $params['share'] = trim($params['share'], '/');
        $this->shareName = $params['share'];
        $this->host = $params['host'];
        $this->user = $params['user'];
        $this->password = $params['password'];
        $this->root = $params['root'];
        $this->credentialsType = $params['credentialsType'];

        $this->getWatcher()->setPolicy(Watcher::CHECK_ONCE);

        /* Register in the singleton notifier */
        $this->wndNotifier = WNDNotifier::getSingleton();
        $this->wndNotifier->registerWND($this);
    }

    private function fixConstructParams($params) {
        $p = $params;
        unset($p['password']);
        $this->log('fixConstructParams: '.print_r($p, true), Util::DEBUG);
        // if user is null we convert username and password to '' to try anonymous login
        $params['user'] = Utils::getDefault($params, 'user', '');
        $params['password'] = Utils::getDefault($params, 'password', '');

        //make sure the root end like "/something/" or "/something/more/"
        $params['root'] = Utils::getDefault($params, 'root', '');
        $params['root'] = '/' . trim($params['root'], '/') . '/';
        //fix special case
        if ($params['root'] === '//') {
            $params['root'] = '/';
        }

        $params['oc_sharing'] = (bool) Utils::getDefault($params, 'oc_sharing', false);

        return $params;
    }

    /**
     * creates instances of NativeServer and fetches the Share instance of it
     *
     * @return bool true if the share has been initialized
     * @throws StorageNotAvailableException
     */
    private function init() {
        $this->initComponents();
        $this->checkConnection();
    }

    private function initComponents() {
        if($this->isInitialized) {
            return true;
        }
        $this->log("enter: initComponents", Util::DEBUG);

        if($this->password === '' && $this->user !== '') {
            $this->log("leave: initComponents StorageNotAvailableException", Util::DEBUG);
            throw new StorageNotAvailableException('Password required when username given');
        }

        $this->server = new NativeServer($this->host, $this->user, $this->password);
        $this->share = $this->server->getShare($this->shareName);
        $this->dirCache = new CappedMemoryCache(5);
        $this->statCache = new CappedMemoryCache(2048);
        $this->xattrCache = new CappedMemoryCache();

        $this->isInitialized = true;
        $this->log("leave: initComponents '{$this->getId()}'", Util::DEBUG);
        return true;
    }

    private function checkConnection() {
        if ($this->connectionTested) {
            return true;
        }
        $this->log("enter: checkConnection", Util::DEBUG);

        $this->connectionTested = true;
        $result = false;
        try {
            $this->test(true);
            $result = true;
        } catch (ForbiddenException $e) {
            if($this->password !== '' || $this->user !== '') {
                // the test will set the isInitialized flag to false if needed
                throw $this->leave(__FUNCTION__, new StorageNotAvailableException('Storage not available, perhaps outdated credentials.'));
            }
        }
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * Get the username trying to access to the windows drive. It might be different than the
     * current ownCloud user
     *
     * @return string the username
     */
    public function getUser() {
        return $this->user;
    }

    /**
     * Get the host we're trying to access
     *
     * @return string the host
     */
    public function getHost() {
        return $this->host;
    }

    /**
     * Get the sharename where we're trying to access
     *
     * @return string the share name
     */
    public function getShareName() {
        return $this->shareName;
    }

    /**
     * Get the folder that will act as our root
     *
     * @return string the root folder
     */
    public function getRoot() {
        return $this->root;
    }

    /**
     * Get the password we'll use to connect to the windows drive. Use it only for unittest
     *
     * @return string the password we'll use
     */
    public function getPassword() {
        $this->log('getPassword', Util::DEBUG);
        return $this->password;
    }

    /**
     * Get the FS id (for ownCloud purposes)
     *
     * @return string the FS id
     */
    public function getId(){
        $id = 'wnd::' . $this->getUser() . '@' . $this->getHost() . '/' . $this->getShareName() . '/' . $this->getRoot();
        return $id;
    }

    private function dir($path) {
        $this->log("enter: dir($path)", Util::DEBUG);
        $this->init();
        $path = trim($path, '/');
        $result = false;
        try {
            if (!$this->dirCache->hasKey($path)) {
                $children = [];
                $dirData = $this->share->dir($path);
                foreach ($dirData as $fileInfo) {
                    // check if the file is readable before adding it to the list
                    // can't use "isReadable" function here, use smb internals instead
                    if (!$fileInfo->isHidden()) {
                        $name = $fileInfo->getName();
                        $children[] = $name;
                        //remember entry so we can answer file_exists and filetype without a full stat
                        $this->statCache->set($path.'/'.$name, $fileInfo);
                    } else {
                        $this->log("{$fileInfo->getName()} isn't readable, skipping", Util::DEBUG);
                    }
                }
                $this->dirCache->set($path, $children);
            }
            $result = $this->dirCache->get($path);
        } catch (NotFoundException $e) {
            //FIXME this smells like fixing symptomps
            $this->swallow(__FUNCTION__, $e);
            $result = [];
        } catch (\Exception $e){
            throw $this->leave(__FUNCTION__, $e);
        }
        return $this->leave(__FUNCTION__, $result);
    }

    private function removeFromCache($path) {
        $path = trim($path, '/');
        // TODO The CappedCache does not really clear by prefix. It just clears all.
        $this->dirCache->clear($path);
        $this->statCache->clear($path);
        $this->xattrCache->clear($path);
    }


    /**
     * Create a directory. Notice that all the parent directories should exists
     *
     * @param string $path the directory we want to create
     * @return bool true if the directory is created, false otherwise
     */
    public function mkdir($path) {
        $this->log("enter: mkdir($path)", Util::DEBUG);
        $this->init();
        $path = trim($path, '/');
        $result = false;
        try {
            if (!$this->file_exists($path)) {
                $this->removeFromCache($this->root . $path);
                $this->share->mkdir($this->root . $path);
                $result = true;
            }
        } catch (\Exception $e){
            $this->swallow(__FUNCTION__, $e);
        }
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * Remove the directory and its contents
     *
     * @param string $path the FS path that we want to delete
     * @return bool true if deleted, false otherwise
     */
    public function rmdir($path) {
        $this->log("enter: rmdir($path)", Util::DEBUG);
        $this->init();
        $path = trim($path, '/');
        $result = false;
        try {
            $content = $this->dir($this->root . $path);
            foreach ($content as $name) {
                $stat = $this->stat($path . '/' . $name);
                if ($stat['type'] === 'dir') {
                    $this->rmdir($path . '/' . $name);
                } else {
                    $this->unlink($path . '/' . $name);
                }
            }
            $this->share->rmdir($this->root . $path);
            $this->removeFromCache($this->root . $path);
            $result = true;
        } catch (\Exception $e){
            $this->swallow(__FUNCTION__, $e);
        }
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * Open the directory
     *
     * @param string $path FS path
     * @return resource a directory resource
     */
    public function opendir($path) {
        $this->log("enter: opendir($path)", Util::DEBUG);
        $this->init();
        $path = trim($path, '/');
        try {
            $files = $this->dir($this->root . $path);
        } catch (\Exception $e){
            $this->swallow(__FUNCTION__, $e);
            $files = [];
        }
        Dir::register('wnd:' . $path, $files);
        return $this->leave(__FUNCTION__, opendir('fakedir://wnd:' . $path));
    }

    /**
     * Get the file type of the path
     *
     * @param string $path FS path
     * @return string the type of the path (usually 'dir' or 'file')
     */
    public function filetype($path) {
        $this->log("enter: filetype ($path)", Util::DEBUG);
        $this->init();
        $path = trim($path, '/');
        $result = false;
        if ($this->isRootDir($path)) {
            $result = 'dir';
        } else {
            //FIXME we should be able to use the minimal stat from a dir() call, es that also returns a type:
            // see https://github.com/eduardok/libsmbclient-php#smbclient_readdir
            // needs to ge int icewind/SMB: https://github.com/icewind1991/SMB/blob/master/src/NativeShare.php#L88
            $stat = $this->stat($path);
            if ($stat) {
                $result = $stat['type'];
            }
        }
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * Check the the file defined by path exists
     *
     * @param string $path FS path
     * @return bool true if the file exists, false otherwise
     */
    public function file_exists($path) {
        $this->log("enter: file_exists($path)", Util::DEBUG);
        $this->init();
        $path = trim($path, '/');
        $result = false;
        if ($this->isRootDir($path) && $this->remoteIsShare()) {
            // referring to a share (not folder) -> return true
            $this->log("leave: file_exists, result: true, its a share ", Util::DEBUG);
            $result = true;
        } else {

            try {
                $result = $this->statCache->hasKey($path);
                if ($result) {
                    $this->log("file_exists, cache hit for '$path'", Util::DEBUG);
                } else {
                    $stat = $this->stat($path);
                    if (is_array($stat)) {
                        $result = true;
                    }
                }
            } catch (NotFoundException $e) {
                $this->swallow(__FUNCTION__, $e);
            } catch (\Exception $e) {
                throw $this->leave(__FUNCTION__, new StorageNotAvailableException($e->getMessage(), $e->getCode(), $e));
            }
        }
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * Remove the path. If the path is a directory, this function falls back to the rmdir
     * function
     *
     * @param string $path FS path to be deleted
     * @return bool true if the path no longer exists, false otherwise
     */
    public function unlink($path) {
        $this->log("enter: unlink($path)", Util::DEBUG);
        $this->init();
        $path = trim($path, '/');
        $result = false;
        try {
            if ($this->is_dir($path)) {
                $result = $this->rmdir($path);
            } else {
                $this->removeFromCache($this->root . $path);
                $result = $this->share->del($this->root . $path);
            }
        } catch (\Exception $e){
            $this->swallow(__FUNCTION__, $e);
        }
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * Open a file resouce
     *
     * @param string $path file path to be opened
     * @param string $mode open the file in this mode
     * @return resource an opened resource to read or write
     * @throws \Exception
     */
    public function fopen($path, $mode) {
        $this->log("enter: fopen ($path, $mode)", Util::DEBUG);
        $this->init();
        $path = trim($path, '/');
        $fullPath = $this->buildPath($path);
        $result = false;
        try {
            switch ($mode) {
                case 'r':
                case 'rb':
                    if (!$this->file_exists($path)) {
                        break;
                    }
                    $result = $this->share->read($fullPath);
                    break;
                case 'w':
                case 'wb':
                    $source = $this->share->write($fullPath);
                    $result = CallbackWrapper::wrap($source, null, null, function () use ($path) {
                        $this->removeFromCache($this->root . $path);
                    });
                    break;
                case 'a':
                case 'ab':
                case 'r+':
                case 'w+':
                case 'wb+':
                case 'a+':
                case 'x':
                case 'x+':
                case 'c':
                case 'c+':
                    //emulate these
                    $tempManager = \OC::$server->getTempManager();
                    if (strrpos($path, '.') !== false) {
                        $ext = substr($path, strrpos($path, '.'));
                    } else {
                        $ext = '';
                    }
                    if ($this->file_exists($path)) {
                        if (!$this->isUpdatable($path)) {
                            break;
                        }
                        if ($mode === 'w' or $mode === 'w+') {
                            $tmpFile = $tempManager->getTemporaryFile($ext);
                        } else {
                            $tmpFile = $this->getCachedFile($path);
                        }
                    } else {
                        if (!$this->isCreatable(dirname($path))) {
                            break;
                        }
                        $tmpFile = $tempManager->getTemporaryFile($ext);
                    }
                    $source = fopen($tmpFile, $mode);
                    $share = $this->share;
                    $result = CallbackWrapper::wrap($source, null, null, function () use ($tmpFile, $fullPath, $path, $share) {
                        $this->removeFromCache($this->root . $path);
                        $share->put($tmpFile, $fullPath);
                        unlink($tmpFile);
                    });
            }
        } catch (Exception $e){
            // swallow smb exceptions, class name will be logged
            $this->swallow(__FUNCTION__, $e);
        } catch (\Exception $e){
            // The rest is critical, so we pass it on
            throw $this->leave(__FUNCTION__, $e);
        }
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * "Touch" the path. This function will only create a empty file if not exists or update
     * the mtime to the latest if it exists. Setting the mtime is not supported
     *
     * @param string $path path to be "touched"
     * @param mixed $time null to be updated to the latest, if it's not null, it will be ignored
     * @return bool true if file exists and mtime updated
     * @throws \Exception
     */
    public function touch($path, $time = null) {
        $this->log("enter: touch ('$path', '$time')", Util::DEBUG);
        $this->init();
        $result = false;
        if (is_null($time) && !$this->file_exists($path)) {
            try{
                $this->removeFromCache($this->root . $path);
                $fh = $this->share->write($this->buildPath($path));
            } catch (\Exception $e){
                throw $this->leave(__FUNCTION__, $e);
            }
            fclose($fh);
            $result = true;
        }
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * Get the mimetype of the path or false if the path not exists
     *
     * @param string $path the target path
     * @return string|false the mimetype or false
     */
    public function getMimeType($path) {
        $this->log("enter: getMimeType($path)", Util::DEBUG);
        $this->init();
        $result = false;
        if ($this->is_dir($path)) {
            $result = 'httpd/unix-directory';
        } else if ($this->file_exists($path)) {
            $result = \OC::$server->getMimeTypeDetector()->detectPath($path);
        }
        return $this->leave(__FUNCTION__, $result);
    }


    /**
     * Rename the files
     *
     * @param string $source the old name of the path
     * @param string $target the new name of the path
     * @return bool true if the rename is successful, false otherwise
     */
    public function rename($source, $target) {
        $this->log("enter: rename('$source', '$target')", Util::DEBUG);
        $this->init();
        try {
            $result = $this->share->rename($this->root . $source, $this->root . $target);
            $this->removeFromCache($this->root . $source);
            $this->removeFromCache($this->root . $target);
        } catch (AlreadyExistsException $e) {
            $this->unlink($target);
            $result = $this->share->rename($this->root . $source, $this->root . $target);
            $this->removeFromCache($this->root . $source);
            $this->removeFromCache($this->root . $target);
            $this->swallow(__FUNCTION__, $e);
        } catch (\Exception $e) {
            $this->swallow(__FUNCTION__, $e);
            $result = false;
        }
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * perform the stat call over the file
     *
     * @param string $path the FS path
     * @return array|false an array with information or false in case of error
     */
    public function stat($path) {
        $this->log("enter: stat($path)", Util::DEBUG);
        $this->init();
        $path = trim($path, '/');
        // The fileinfo objects may throw exceptions when calling getters
        try {
            if ($this->remoteIsShare() && $this->isRootDir($path)) { //mtime doesn't work for shares
                $this->log("stat '$path' is remoteIsShare", Util::DEBUG);
                $result = [
                    'mtime' => $this->shareMTime(),
                    'size' => 0,
                    'type' => 'dir'
                ];
            } else {
                $stat = $this->statCache->get($path);
                if ($stat) {
                    $this->log("stat cache hit for '$path'", Util::DEBUG);
                    $result = [
                        'mtime' => $stat->getMTime(),
                        'size' => $stat->getSize(),
                        'type' => ($stat->isDirectory()) ? 'dir' : 'file'
                    ];
                } else {
                    $this->log("stat fetching '{$this->root}$path'", Util::DEBUG);
                    /** @var \Icewind\SMB\NativeFileInfo[] $fileInfo */
                    $fileInfo = $this->share->stat($this->root . $path);
                    $this->statCache->set($path, $fileInfo);
                    $result = [
                        'mtime' => $fileInfo->getMTime(),
                        'size' => $fileInfo->getSize(),
                        'type' => ($fileInfo->isDirectory()) ? 'dir' : 'file'
                    ];
                }
            }
        } catch (\Exception $e){
            $this->swallow(__FUNCTION__, $e);
            $result = false;
        }
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * Check the path has updated in the given time. The check is usually based on the mtime
     * of the path, so if the path's filemtime is greater than $time, the file has updated
     *
     * @param string $path the FS path
     * @param int $time timestamp to compare against
     * @throw \OCP\Files\StorageNotAvailableException
     * @return bool true if the file has updated, false otherwise
     * @throws StorageNotAvailableException
     */
    public function hasUpdated($path, $time) {
        $this->log('enter: '.__FUNCTION__."('$path', '$time')", Util::DEBUG);
        $this->init();
        $stat = $this->stat($path);
        if ($stat === false) {
            //FIXME ... quite a claim, maybe the file was just not found
            Util::writeLog('wnd', 'hasUpdated failed -> storage not available', Util::ERROR);
            throw $this->leave(__FUNCTION__, new StorageNotAvailableException());
            // if no stat consider a the file has updated to force a rescan
            //return true;
        } else {
            $currentMtime = $stat['mtime'];
            $result = $currentMtime > $time;
        }
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * get the best guess for the modification time of the share
     *
     * @return array the first element of the array is the calculated mtime for the folder, and
     * the second is the list of readable files (the list contains the filename as key and the
     * mtime for the file as value)
     */
    private function shareMTime() {
        $this->log('enter: '.__FUNCTION__, Util::DEBUG);
        $files = $this->dir($this->root);
        $result = 0;
        foreach ($files as $file) {
            $stat = $this->stat($file);
            if ($stat['mtime'] > $result) {
                $result = $stat['mtime'];
            }
        }
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * Check if the path is our root dir (not the smb one)
     *
     * @param string $path the path
     * @return bool true if it's root, false if not
     */
    private function isRootDir($path) {
        $this->log('enter: '.__FUNCTION__."($path)", Util::DEBUG);
        $result = $path == '' || $path == '/' || $path == '.';
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * Check if our root points to a smb share
     *
     * @return bool true if our root points to a share false otherwise
     */
    private function remoteIsShare() {
        $result = $this->getShareName() && (!$this->getRoot() || $this->getRoot() === '/');
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * Check if our root points to a smb path
     *
     * @return bool true if our root points to a path false otherwise
     */
    private function remoteIsFolder() {
        $result = $this->getShareName() && ($this->getRoot() && $this->getRoot() !== '/');
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * Check where our root points to remotely. It can be a smb share ("share") or one of the
     * inner folder ("folder")
     *
     * @return string|false 'share' or 'folder'
     */
    public function remoteType() {
        $this->log('enter: remoteType', Util::DEBUG);
        if ($this->remoteIsShare()) {
            $result = 'share';
        } else if ($this->remoteIsFolder()) {
            $result = 'folder';
        } else {
            //FIXME else what?
            $this->log('remoteType encountered unexpected type', Util::ERROR);
            $result = false;
        }
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * Check if the path is readable
     *
     * @param string $path the path
     * @return bool true if it's readable, false if not
     */
    public function isReadable($path) {
        $this->log("enter: isReadable($path)", Util::DEBUG);
        $this->init();
        if ($this->isRootDir($path)) {
            $this->log("isReadable path: root is always readable", Util::DEBUG);
            $result = true;
        } else {
            $info = $this->getFileInfoOperation($path, 'isHidden');
            if ($info === null) {
                $result = false;
            } else {
                $result = !$info;
            }
        }
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * Check if the path is updatable
     *
     * @param string $path the path
     * @return bool true if it's updatable, false if not
     */
    public function isUpdatable($path) {
        $this->log("enter: isUpdatable($path)", Util::DEBUG);
        $this->init();
        $info = $this->getFileInfoOperation($path, 'isReadOnly');
        if ($info === null) {
            $result = false;
        } else {
            $result = !$info;
        }
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * Check if the path is creatable
     *
     * @param string $path the path
     * @return bool true if it's creatable, false if not
     */
    public function isCreatable($path) {
        $this->log("enter: isCreatable($path)", Util::DEBUG);
        $this->init();
        if (!$this->isRootDir($path)) {
            // if the path isn't root, use the default behaviour
            $result = parent::isCreatable($path);
        } else {
            // if path is root check that the folder isn't read-only
            $info = $this->getFileInfoOperation($path, 'isReadOnly');
            if ($info === null) {
                $result = false;
            } else {
                $result = !$info;
            }
        }
        return $this->leave(__FUNCTION__, $result);
    }

    private function getFileInfoOperation($path, $operation) {
        $this->log("enter: getFileInfoOperation('$path', '$operation')", Util::DEBUG);
        $this->init();
        if ($this->xattrCache->hasKey($path)) {
             $nativeFI = $this->xattrCache->get($path);
        } else {
             $realPath = $this->root . $path;
             $nativeFI = new NativeFileInfo($this->share, $realPath, basename($realPath));
             $this->xattrCache->set($path, $nativeFI);
        }

        try {
            $result = call_user_func(array($nativeFI, $operation));
        } catch (\Exception $e){
            $this->swallow(__FUNCTION__, $e);
            $result = null;
        }
        return $this->leave(__FUNCTION__, $result);
    }

    /**
     * Check if the path is sharable. Currently sharing is disabled so it will always
     * return false
     *
     * @param string $path the path
     * @return false
     */
    public function isSharable($path) {
        $this->log("enter: isSharable($path)", Util::DEBUG);
        return $this->isSharingActive;
    }

    /**
     * @param string $path
     * @return string
     */
    protected function buildPath($path) {
        $this->log("enter: buildPath($path)", Util::DEBUG);
        return Filesystem::normalizePath($this->root . '/' . $path);
    }

    public function receiveNotificationFrom(WND $other, $changeType) {
        $this->log("enter: receiveNotificationFrom '{$other->getHost()}' '{$other->getUser()}' to '{$this->getHost()}' '{$this->getUser()}', $changeType", \OCP\Util::WARN);
        if ($changeType === WNDNotifier::NOTIFY_PASSWORD_REMOVAL &&
                $this->getHost() === $other->getHost() &&
                $this->getUser() === $other->getUser()) {
            $this->password = '';
            $this->isInitialized = false;
        }
    }

    /**
     * @param string $message
     * @param int $level
     * @param string $from
     */
    private function log($message, $level, $from = 'wnd') {
        if(\OC::$server->getConfig()->getSystemValue('wnd.logging.enable', false) === true){
            Util::writeLog($from, $message, $level);
        }
    }

    /**
     * A simple function to test connectivity. This function might throw exceptions.
     * By default, the function won't reset the password set for the connection. This is useful
     * for simple tests with new _possible_ parameters (connection tests performed before saving
     * the parameters in the DB)
     * For real connections (the ones the user want to use), it's possible to reset the password.
     *
     * NOTE: resetting the password will wipe out all the passwords set for those mounts whose host,
     * share and user WND account match the current one. This could potentially remove passwords for
     * other users, and / or being admin or personal mounts, but this shouldn't be a problem because
     * all those users would be accessing using the same *wrong* account
     *
     * @param bool $resetPassword
     * @return bool
     * @throws ForbiddenException
     */
    public function test($resetPassword = false) {
        $this->log("enter: test($resetPassword)", Util::DEBUG);
        $this->initComponents();
        try {
            //always check remote directly
            $this->share->dir(trim($this->root, '/'));
        } catch (ForbiddenException $ex) {
            if ($resetPassword && $this->password !== '' &&
                    !in_array($this->credentialsType,
                              array(WNDUtils::CUSTOM_CREDENTIALS_ADMIN_LOGIN),
                              true)) {
                // don't bother to reset passwords if this one is empty
                // and trigger the reset if the credentials aren't global or login ones
                WNDUtils::resetPassword($this);
                // notify all WND objects to reset the password if needed
                // this object will also be notified
                $this->wndNotifier->notifyChange($this, WNDNotifier::NOTIFY_PASSWORD_REMOVAL);
            }
            // TODO we should probably throw an \OC\ForbiddenException exception?
            throw $this->leave(__FUNCTION__, $ex);
        } catch (\Exception $ex) {
            throw $this->leave(__FUNCTION__, $ex);
        }
        // mark the storage as available, so the availability wrapper is up to date
        $this->setAvailability(true);
        return $this->leave(__FUNCTION__, true);
    }

    public function getMetaData($path) {
        $this->log("enter: getMetadata($path)", Util::DEBUG);
        try {
            $this->init();
            return parent::getMetaData($path);
        } catch (StorageNotAvailableException $e) {
            if ($this->isRootDir($path)) {
                return array('mimetype' => 'httpd/unix-directory',
                             'mtime' => 0,
                             'size' => -1,
                             'etag' => $this->getETag($path),
                             'storage_mtime' => 0,
                             'permission' => \OCP\Constants::PERMISSION_READ);
            } else {
                return null;
            }
        }
    }

    public function remove($path) {
        $this->init();
        return parent::remove($path);
    }

    public function is_dir($path) {
        $this->init();
        return parent::is_dir($path);
    }

    public function is_file($path) {
        $this->init();
        return parent::is_file($path);
    }

    public function filesize($path) {
        $this->init();
        return parent::filesize($path);
    }

    public function isDeletable($path) {
        $this->init();
        return parent::isDeletable($path);
    }

    public function getPermissions($path) {
        $this->init();
        return parent::getPermissions($path);
    }

    public function filemtime($path) {
        $this->init();
        return parent::filemtime($path);
    }

    public function file_get_contents($path) {
        $this->init();
        return parent::file_get_contents($path);
    }

    public function file_put_contents($path, $data) {
        $this->init();
        return parent::file_put_contents($path, $data);
    }

    public function copy($path1, $path2) {
        $this->init();
        return parent::copy($path1, $path2);
    }

    public function hash($type, $path, $raw = false) {
        $this->init();
        return parent::hash($type, $path, $raw);
    }

    public function searchInDir($query, $dir = '') {
        $this->init();
        return parent::searchInDir($query, $dir);
    }

    public function getLocalFolder($path) {
        $this->init();
        return parent::getLocalFolder($path);
    }

    /**
     * if wnd.logging.enable is set to true in the config will log a leave line
     * with the given function, the return value or the exception
     *
     * @param $function
     * @param mixed $result an exception will be logged and then returned
     * @return mixed
     */
    private function leave($function, $result) {
        if(\OC::$server->getConfig()->getSystemValue('wnd.logging.enable', false) === false){
            //don't bother building log strings
            return $result;
        } else if ($result === true) {
            $this->log("leave: $function, return true", Util::DEBUG);
        } else if ($result === false) {
            $this->log("leave: $function, return false", Util::DEBUG);
        } else if (is_string($result)) {
            $this->log("leave: $function, return '$result'", Util::DEBUG);
        } else if (is_resource($result)) {
            $this->log("leave: $function, return resource", Util::DEBUG);
        } else if ($result instanceof \Exception) {
            $this->log("leave: $function, throw ".get_class($result).' - code: '.$result->getCode().' message: '.$result->getMessage().' trace: '.$result->getTraceAsString(), Util::DEBUG);
        } else {
            $this->log("leave: $function, return ".json_encode($result), Util::DEBUG);
        }
        return $result;
    }

    private function swallow($function, \Exception $exception) {
        if(\OC::$server->getConfig()->getSystemValue('wnd.logging.enable', false) === true){
            $this->log("$function swallowing ".get_class($exception).' - code: '.$exception->getCode().' message: '.$exception->getMessage().' trace: '.$exception->getTraceAsString(), Util::DEBUG);
        }
    }
}

