<?php
/**
 * ownCloud
 *
 * @author Juan Pablo Villafañez Ramos <jvillafanez@owncloud.com>
 * @author Jesus Macias Portela <jesus@owncloud.com>
 * @copyright (C) 2014 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\windows_network_drive\lib;

use \OCA\windows_network_drive\lib\Utils;
use \OCA\windows_network_drive\lib\DBAccess;
use OCP\Files\StorageNotAvailableException;

/**
 * Utility class to manage WND easily.
 */
class WNDUtils {
    // these have **nothing** to do as with what is stored in the DB as use_custom_credentials
    const CUSTOM_CREDENTIALS_ADMIN_CUSTOM        = 1;
    const CUSTOM_CREDENTIALS_ADMIN_USER          = 2;
    const CUSTOM_CREDENTIALS_ADMIN_GLOBAL        = 3;
    const CUSTOM_CREDENTIALS_ADMIN_LOGIN         = 4;
    const CUSTOM_CREDENTIALS_USER_CUSTOM         = 5;
    const CUSTOM_CREDENTIALS_USER_GLOBAL         = 6;

    private static function decryptPasswordInParams($fsparams) {
        if (!array_key_exists('password', $fsparams)) {
            \OCP\Util::writeLog('wnd', 'missing password mounting the FS', \OCP\Util::ERROR);
            return $fsparams;
        }

        if ($fsparams['password'] !== null) {
            $correctPassword = Utils::decryptPassword($fsparams['password']);
            if ($correctPassword === false) {
                $logMessage = 'Could not decrypt credentials in parameters';
                $logMessage .= (isset($fsparams['user']) && $fsparams['user'] !== '') ? " for \"{$fsparams['user']}\"" : '';
                $logMessage .= (isset($fsparams['host']) && $fsparams['host'] !== '') ? " on \"{$fsparams['host']}\"" : '';
                \OCP\Util::writeLog('wnd', $logMessage, \OCP\Util::ERROR);
                return $fsparams;
            }

            $fsparams['password'] = $correctPassword;
        }
        return $fsparams;
    }
    private static function mountWndWithParams($user, $rowMountPoint, array $fsparams) {
        if(!isset($fsparams['oc_sharing'])){
            $fsparams['oc_sharing'] = \OCP\Config::getAppValue('windows_network_drive', 'global_sharing', false);
        }
        $fsparams["share"] = str_replace('$user', \OC_User::getUser(), $fsparams["share"]);
        $fsparams["root"] = str_replace('$user', \OC_User::getUser(), $fsparams["root"]);
        $mountPoint = "/$user/files/" . $rowMountPoint;
        \OC\Files\Filesystem::mount('\OCA\windows_network_drive\lib\WND',
                                    $fsparams, $mountPoint);
    }

    /**
     * compiles and returns parameters for the storage, based on database data
     *
     * @param array $row
     * @param null|string $user
     * @return array
     */
    public static function getParamsForGlobalRow($row, $user = null) {
        if(is_null($user) && isset($row['mounting_user'])) {
            $user = $row['mounting_user'];
        }
        $fsparams = array(
            'host'       => $row['mount_url'],
            'share'      => $row['share'],
            'oc_sharing' => \OCP\Config::getAppValue('windows_network_drive', 'global_sharing', false),
            'root'       => $row['root'],
            'user'       => null,
            'password'   => null,
        );

        switch(intval($row['use_custom_credentials'])) {
            case 0:
                if(!is_null($row['global_username']) && !is_null($row['global_password'])) {
                    $fsparams['user']     = $row['global_username'];
                    $fsparams['password'] = $row['global_password'];
                }
                $fsparams['credentialsType'] = WNDUtils::CUSTOM_CREDENTIALS_ADMIN_CUSTOM;
                break;

            case 1:
                $fsparams['credentialsType'] = WNDUtils::CUSTOM_CREDENTIALS_ADMIN_USER;
                if(!is_null($row['credential_username']) && !is_null($row['credential_password'])) {
                    if(intval($row['use_global_credentials']) === 0) {
                        $fsparams['user'] = $row['credential_username'];
                        $fsparams['password'] = $row['credential_password'];
                        $fsparams['credentialsType'] = WNDUtils::CUSTOM_CREDENTIALS_ADMIN_USER;
                    } else {
                        // This is not possible at the moment...
                        $userGlobalCreds = DBAccess::getPersonalGlobalCredentialsCached($user);
                        $fsparams['user'] = $userGlobalCreds['global_username'];
                        $fsparams['password'] = $userGlobalCreds['global_password'];
                        $fsparams['credentialsType'] = WNDUtils::CUSTOM_CREDENTIALS_USER_GLOBAL;
                    }
                }
                break;

            case 2:
                $globalCreds = DBAccess::getAdminGlobalCredentialsCached();
                $fsparams['user'] = $globalCreds['global_username'];
                $fsparams['password'] = $globalCreds['global_password'];
                $fsparams['credentialsType'] = WNDUtils::CUSTOM_CREDENTIALS_ADMIN_GLOBAL;
                break;

            case 3:
                $encryptedCredentials = \OC::$server->getSession()->get('wnd-credentials');
                $encodedCredentials = \OC::$server->getCrypto()->decrypt($encryptedCredentials);
                $paramsAuth = json_decode($encodedCredentials, true);
                $fsparams['user'] = $paramsAuth['uid'];
                $fsparams['password'] = $paramsAuth['password'];
                $fsparams['oc_sharing'] = false;
                $fsparams['credentialsType'] = WNDUtils::CUSTOM_CREDENTIALS_ADMIN_LOGIN;

                //Adding info about domain if exist (Note that we use global_username
                //column to store domain info when use this auth type)
                if($row['global_username'] !== '') {
                    // username with domain = 'domain\username'
                    $fsparams["user"] = $row['global_username'].'\\'.$fsparams["user"];
                }
                break;
        }

        $fsparams = self::decryptPasswordInParams($fsparams);
        $fsparams["share"] = str_replace('$user', \OC_User::getUser(), $fsparams["share"]);
        $fsparams["root"]  = str_replace('$user', \OC_User::getUser(), $fsparams["root"]);

        return $fsparams;
    }

    /**
     * Mount a global (admin) mount point for the user with the row data
     *
     * @param string $user the user mounting the storage. This user will be used to get the
     * personal credentials if needed, and to mount the storage. $user should match the
     * $row['mounting_user'] parameter
     * @param array $row a row coming from the DB with all the data to mount the storage
     */
    public static function mountGlobalWndRow($user, $row) {
        $parameters = self::getParamsForGlobalRow($row, $user);
        self::mountWndWithParams($user, $row['mount_point'], $parameters);
    }

    private static function getParamsForPersonalRow($user, $row) {
        if (intval($row['use_global_credentials']) === 0) {
            $fsparams = array ("host" => $row['mount_url'],
                                "user" => $row['creds_username'],
                                "password" => $row['creds_password'],
                                "share" => $row['share'],
                                "root" => $row['root'],
                                "credentialsType" => WNDUtils::CUSTOM_CREDENTIALS_USER_CUSTOM);
        } else {
            $userGlobalCreds = DBAccess::getPersonalGlobalCredentialsCached($user);
            $fsparams = array ("host" => $row['mount_url'],
                                "user" => $userGlobalCreds['global_username'],
                                "password" => $userGlobalCreds['global_password'],
                                "share" => $row['share'],
                                "root" => $row['root'],
                                "credentialsType" => WNDUtils::CUSTOM_CREDENTIALS_USER_GLOBAL);
        }
        $fsparams = self::decryptPasswordInParams($fsparams);
        return $fsparams;
    }

    /**
     * Mount a personal mount point for the user with the row data
     *
     * @param string $user the user mounting the storage. This user will be used to get the
     * personal credentials if needed, and to mount the storage. $user should match the
     * $row['target_user'] parameter
     * @param array $row a row coming from the DB with all the data to mount the storage
     */
    public static function mountPersonalWndRow($user, $row) {
        $fsparams = self::getParamsForPersonalRow($user, $row);
        self::mountWndWithParams($user, $row['mount_point'], $fsparams);
    }


    /**
     * Get a WND instance according to the data coming from row.
     *
     * @param array $row a row coming from the DB with all the data to mount the global (admin)
     * storage
     * @return \OCA\windows_network_drive\lib\WND The WND instance for the row
     */
    public static function getWndForGlobalRow($row) {
        $parameters = self::getParamsForGlobalRow($row);
        return new \OCA\windows_network_drive\lib\WND($parameters);
    }

    /**
     * Get a WND instance according to the data coming from row.
     *
     * @param array $row a row coming from the DB with all the data to mount the personal storage
     * @return \OCA\windows_network_drive\lib\WND The WND instance for the row
     */
    public static function getWndForPersonalRow($row) {
        $fsparams = self::getParamsForPersonalRow($row['target_user'], $row);
        $fsparams['oc_sharing'] = \OCP\Config::getAppValue('windows_network_drive', 'global_sharing', false);
        return new \OCA\windows_network_drive\lib\WND($fsparams);
    }

    /**
     * reset user-custom per-storage credentials if log in failed
     * in order to avoid user lock-out in AD
     *
     * @param WND $wnd
     * @throws \Exception
     */
    public static function resetPassword(WND $wnd) {
        $user = $wnd->getUser();
        $pwd = $wnd->getPassword();
        if(is_null($user) || $user === '' || is_null($pwd) || $pwd === '') {
            // anonymous login -> nothing to do here
            return;
        }
        DBAccess::wipeOutPasswordForBackend($wnd->getHost(), $wnd->getUser());

    }

    /**
     * Perform a test for the wnd and get results
     *
     * The result is an array with the following information:
     * - "status" : "ok" or "fail"
     * - "code" : if the status is "ok" this field is missing, otherwise it can have the
     * following values
     *  - "TF" : for "test failed"
     *  - "AD" : for "access denied"
     *  - "IH" : for "invalid host"
     *  - "CE" : for "connection error"
     *  - "GE" : for "generic error"
     * - "error" : if the status is "ok" this field is missing, otherwise a translatable error
     * message explaining the situation is here
     *
     * @param \OCA\windows_network_drive\lib\WND $wnd the WND instance where the test will be
     * performed
     * @param string $mountPoint the mount point for the instance
     * @param array $test an array where the results will be stored. As long as the mount points
     * are different, this array can be reused to store more tests
     * @param \OC_L10N $l, and L10N instance to get the translation
     */
    public static function performTestOverWnd($wnd, $mountPoint, &$test, $l) {
        try {
            $result = $wnd->test(true);
            if ($result) {
                $test[$mountPoint] = array('status' => 'ok');
            } else {
                $test[$mountPoint] = array('code' => 'TF', 'error' => $l->t('File doesn\'t exists'));
            }
        } catch (\Icewind\SMB\Exception\AccessDeniedException $e) {
            $test[$mountPoint] = array('code' => 'AD', 'error' => $l->t('Access denied to the server'));
        } catch (\Icewind\SMB\Exception\ForbiddenException $e) {
            $test[$mountPoint] = array('code' => 'AD', 'error' => $l->t('Access denied to the server'));
        } catch (\Icewind\SMB\Exception\InvalidHostException $e) {
            $test[$mountPoint] = array('code' => 'IH', 'error' => $l->t('Invalid host'));
        } catch (\Icewind\SMB\Exception\ConnectionError $e) {
            $test[$mountPoint] = array('code' => 'CE', 'error' => $l->t('Cannot connect to the server'));
        } catch (StorageNotAvailableException $e) {
            $test[$mountPoint] = array('code' => 'AD', 'error' => $e->getMessage());
        } catch (\Exception $e) {
            // handle special cases that don't throw the right exceptions
            $errorCode = $e->getCode();
            switch ($errorCode) {
                case 110:
                case 111:
                case 112:
                case 113:
                    $test[$mountPoint] = array('code' => 'CE', 'error' => $l->t('Cannot connect to the server'));
                    break;
                default:
                    $test[$mountPoint] = array('code' => 'GE', 'error' => $l->t('Unknown error: ' . $e->getMessage()));
            }
        }

        if (!isset($test[$mountPoint]['status'])) {
            $test[$mountPoint]['status'] = 'fail';
        }
    }

    public static function performTestOverGlobalRow($row, &$test, $l) {
        $wnd = WNDUtils::getWndForGlobalRow($row);
        if (is_null($wnd)) {
            $test[$row['mount_point']]['code'] = 'CNP';
            $test[$row['mount_point']]['type'] = 'global';
            $test[$row['mount_point']]['location'] = intval($row['use_custom_credentials']);
            $test[$row['mount_point']]['mid'] = intval($row['mount_id']);
            $test[$row['mount_point']]['url'] = $row['mount_url'];
            $test[$row['mount_point']]['share'] = $row['share'];
            $test[$row['mount_point']]['status'] = 'fail';
            $test[$row['mount_point']]['error'] = $l->t('Cannot decrypt credentials');
        } else {
            WNDUtils::performTestOverWnd($wnd, $row['mount_point'], $test, $l);
            // we'll add more information to the test
            $test[$row['mount_point']]['type'] = 'global';
            $test[$row['mount_point']]['location'] = intval($row['use_custom_credentials']);
            $test[$row['mount_point']]['mid'] = intval($row['mount_id']);
            $test[$row['mount_point']]['url'] = $row['mount_url'];
            $test[$row['mount_point']]['share'] = $row['share'];

            if ($test[$row['mount_point']]['status'] === 'ok' ||
                    !in_array($test[$row['mount_point']]['code'], array('CE', 'GE'))) {
                // if check is ok or the error is not due to not having access to the server
                // we need to check if we have valid credentials, to show the proper message
                if ($row['use_custom_credentials'] === "1" && (is_null($row['credential_username']) || is_null($row['credential_password']))) {
                    $test[$row['mount_point']] = array('status' => 'fail', 'mid' => intval($row['mount_id']),
                                                        'location' => intval($row['use_custom_credentials']),
                                                        'type' => 'global', 'code' => 'CNP',
                                                        'url' => $row['mount_url'],
                                                        'share' => $row['share'],
                                                        'error' => $l->t('Credentials not provided yet'));
                }
            }
        }
    }

    public static function performTestOverPersonalRow($row, &$test, $l) {
        $wnd = WNDUtils::getWndForPersonalRow($row);
        if (is_null($wnd)) {
            $test[$row['mount_point']]['code'] = 'CNP';
            $test[$row['mount_point']]['type'] = 'personal';
            $test[$row['mount_point']]['location'] = intval($row['use_global_credentials']);
            $test[$row['mount_point']]['mid'] = intval($row['personal_mount_id']);
            $test[$row['mount_point']]['url'] = $row['mount_url'];
            $test[$row['mount_point']]['share'] = $row['share'];
            $test[$row['mount_point']]['status'] = 'fail';
            $test[$row['mount_point']]['error'] = $l->t('Cannot decrypt credentials');
        } else {
            WNDUtils::performTestOverWnd($wnd, $row['mount_point'], $test, $l);
            $test[$row['mount_point']]['type'] = 'personal';
            $test[$row['mount_point']]['location'] = intval($row['use_global_credentials']);
            $test[$row['mount_point']]['mid'] = intval($row['personal_mount_id']);
            $test[$row['mount_point']]['url'] = $row['mount_url'];
            $test[$row['mount_point']]['share'] = $row['share'];
        }
    }

    public static function getMountsForApi($params) {
        $entries = array();
        $user = \OC_User::getUser();
        $groups = \OC_Group::getUserGroups($user);
        $mountList = DBAccess::getMountsForUser($user, $groups);
        if ($mountList !== false) {
            while (($row = $mountList->fetchRow()) !== false) {
                $entries[] = self::generateApiEntryForRow($row, 'system');
            }
        }

        $personalMountList = DBAccess::getPersonalMountsPerUser($user);
        if ($personalMountList !== false) {
            while (($row = $personalMountList->fetchRow()) !== false) {
                $entries[] = self::generateApiEntryForRow($row, 'personal');
            }
        }
        return new \OC_OCS_Result($entries);
    }

    private static function generateApiEntryForRow($row, $type) {
        $l = \OC_L10N::get('windows_network_drive');  //TODO: Need to find a better place for this
        $path = dirname($row['mount_point']);
        if ($path === '.') {
            $path = "";
        }
        $map = array('personal' => array(0 => $l->t('Credentials only for this mount'),
                                        2 => $l->t('Personal credentials')),
                    'system' => array(0 => $l->t('Admin-provided credentials'),
                                        1 => $l->t('Per-user credentials'),
                                        2 => $l->t('Admin-provided credentials'),
                                        3 => $l->t('Login credentials')));
        $columnKey = null;
        if ($type === 'system') {
            $columnKey = 'use_custom_credentials';
        } else if ($type === 'personal') {
            $columnKey = 'use_global_credentials';
        }

        if ($columnKey === null) {
            $auth = $l->t('Unknown');
        } else {
            $auth = Utils::getDefault($map[$type], $row[$columnKey], $l->t('Unknown'));
        }

        return array(
            'name' => basename($row['mount_point']),
            'server' => $row['mount_url'],
            'share' => $row['share'],
            'auth' => $auth,
            'path' => $path,
            'type' => 'dir',
            'scope' => $type,
            'permissions' => \OCP\PERMISSION_READ
        );
    }
}
